from .views import APIView
from .generics import GenericAPIView
from . import mixins


class ViewSetMixin:
    """
    Overrides `.as_view()` so that it takes an `actions` keyword that performs
    the binding of HTTP methods to actions on the Resource.

    For example, to create a concrete view binding the 'GET' and 'POST' methods
    to the 'list' and 'create' actions...

    view = MyViewSet.as_view({'get': 'list', 'post': 'create'})
    """

    @classmethod
    def as_view(cls, action_map=None, **initkwargs):
        """
        Allows custom request to method routing based on given action_map.
        Needs to re-implement the method but contains all the things the parent does.
        """

        if not action_map:  # actions must not be empty
            raise TypeError("action_map is a required argument.")

        def view(request):
            self = cls(**initkwargs)
            self.request = request
            self.lookup_url_kwargs = self.request.matchdict
            self.action_map = action_map
            self.action = self.action_map.get(self.request.method.lower())

            for method, action in action_map.items():
                handler = getattr(self, action)
                setattr(self, method, handler)

            return self.dispatch(self.request, **self.request.matchdict)

        return view


class APIViewSet(ViewSetMixin, APIView):
    """
    Does not provide any actions by default.
    """
    pass


class GenericAPIViewSet(ViewSetMixin, GenericAPIView):
    """
    The GenericAPIView class does not provide any actions by default,
    but does include the base set of generic view behavior, such as
    the `get_object` and `get_queryset` methods.
    """
    pass


class ReadOnlyModelViewSet(mixins.RetrieveModelMixin,
                           mixins.ListModelMixin,
                           GenericAPIViewSet):
    """
    A viewset that provides default `list()` and `retrieve()` actions.
    """
    pass


class ModelCRUDViewSet(mixins.CreateModelMixin,
                       mixins.RetrieveModelMixin,
                       mixins.UpdateModelMixin,
                       mixins.DestroyModelMixin,
                       mixins.ListModelMixin,
                       GenericAPIViewSet):
    """
    A viewset that provides default `create()`, `retrieve()`, `update()`, `destroy()` and `list()` actions.
    """
    pass


class ModelCRPDViewSet(mixins.CreateModelMixin,
                       mixins.RetrieveModelMixin,
                       mixins.PartialUpdateMixin,
                       mixins.DestroyModelMixin,
                       mixins.ListModelMixin,
                       GenericAPIViewSet):
    """
    A viewset that provides default `create()`, `retrieve()`, `partial_update()`, `destroy()` and `list()` actions.
    """
    pass


class ModelCRUPDViewSet(mixins.CreateModelMixin,
                        mixins.RetrieveModelMixin,
                        mixins.UpdateModelMixin,
                        mixins.PartialUpdateMixin,
                        mixins.DestroyModelMixin,
                        mixins.ListModelMixin,
                        GenericAPIViewSet):
    """
    A viewset that provides default `create()`, `retrieve()`, `partial_update()`, `'update()`,
    `destroy()` and `list()` actions.
    """
    pass



import itertools

from collections import namedtuple

from .exceptions import ImproperlyConfigured

__all__ = ['ViewSetRouter']

Route = namedtuple('Route', ['url', 'mapping', 'name', 'initkwargs'])
DynamicDetailRoute = namedtuple('DynamicDetailRoute', ['url', 'name', 'initkwargs'])
DynamicListRoute = namedtuple('DynamicListRoute', ['url', 'name', 'initkwargs'])


def replace_methodname(format_string, methodname):
    """
    Partially format a format_string, swapping out any '{methodname}' or '{methodnamehyphen}' components.
    """

    methodnamehyphen = methodname.replace('_', '-')
    ret = format_string
    ret = ret.replace('{methodname}', methodname)
    ret = ret.replace('{methodnamehyphen}', methodnamehyphen)
    return ret


def flatten(list_of_lists):
    """
    Takes an iterable of iterables, returns a single iterable containing all items
    """

    return itertools.chain(*list_of_lists)


class ViewSetRouter:
    """
    Automatically adds routes and associates views to the Pyramid Configurator for ViewSets, including
    any decorated list_routes and detail_routes.
    """

    routes = [
        # List route.
        Route(
            url=r'/{prefix}{trailing_slash}',
            mapping={
                'get': 'list',
                'post': 'create'
            },
            name='{basename}-list',
            initkwargs=dict()
        ),
        # Dynamically generated list routes. Generated using @list_route decorator on methods of the viewset.
        DynamicListRoute(
            url=r'/{prefix}/{methodname}{trailing_slash}',
            name='{basename}-{methodnamehyphen}',
            initkwargs={}
        ),
        # Detail route.
        Route(
            url=r'/{prefix}/{lookup}{trailing_slash}',
            mapping={
                'get': 'retrieve',
                'put': 'update',
                'patch': 'partial_update',
                'delete': 'destroy'
            },
            name='{basename}-detail',
            initkwargs=dict()
        ),
        # Dynamically generated detail routes. Generated using @detail_route decorator on methods of the viewset.
        DynamicDetailRoute(
            url=r'/{prefix}/{lookup}/{methodname}{trailing_slash}',
            name='{basename}-{methodnamehyphen}',
            initkwargs={}
        ),
    ]

    def __init__(self, configurator, trailing_slash=True):
        """
        :param configurator: pyramid Configurator
        :return: void
        """

        self.configurator = configurator
        self.trailing_slash = trailing_slash and '/' or ''
        self.registry = list()

    def register(self, prefix, viewset, basename, factory=None, permission=None):
        """
        Factory and permission are likely only going to exist until I have enough time
        to write a permissions module for PRF.

        :param prefix: the uri route prefix.
        :param viewset: The ViewSet class to route.
        :param basename: Used to name the route in pyramid.
        :param factory: Optional, root factory to be used as the context to the route.
        :param permission: Optional, permission to assign the route.
        """

        lookup = self.get_lookup(viewset)
        routes = self.get_routes(viewset)

        for route in routes:
            # Only actions which actually exist on the viewset will be bound
            mapping = self.get_method_map(viewset, route.mapping)

            if not mapping:
                continue  # empty viewset

            url = route.url.format(
                prefix=prefix,
                lookup=lookup,
                trailing_slash=self.trailing_slash
            )
            view = viewset.as_view(mapping, **route.initkwargs)
            name = route.name.format(basename=basename)

            if factory:
                self.configurator.add_route(name, url, factory=factory)
            else:
                self.configurator.add_route(name, url)

            self.configurator.add_view(view, route_name=name, permission=permission)

    def get_routes(self, viewset):
        """
        Augment `self.routes` with any dynamically generated routes.
        Returns a list of the Route namedtuple.
        """

        known_actions = list(flatten([route.mapping.values() for route in self.routes if isinstance(route, Route)]))

        # Determine any `@detail_route` or `@list_route` decorated methods on the viewset
        detail_routes = []
        list_routes = []

        for methodname in dir(viewset):
            attr = getattr(viewset, methodname)
            httpmethods = getattr(attr, 'bind_to_methods', None)
            detail = getattr(attr, 'detail', True)

            if httpmethods:
                # check against know actions list
                if methodname in known_actions:
                    raise ImproperlyConfigured('Cannot use @detail_route or @list_route '
                                               'decorators on method "%s" '
                                               'as it is an existing route'.format(methodname))

                httpmethods = [method.lower() for method in httpmethods]

                if detail:
                    detail_routes.append((httpmethods, methodname))
                else:
                    list_routes.append((httpmethods, methodname))

        def _get_dynamic_routes(route, dynamic_routes):
            ret = []

            for httpmethods, methodname in dynamic_routes:
                method_kwargs = getattr(viewset, methodname).kwargs
                initkwargs = route.initkwargs.copy()
                initkwargs.update(method_kwargs)
                url_path = initkwargs.pop("url_path", None) or methodname

                ret.append(Route(
                    url=replace_methodname(route.url, url_path),
                    mapping={httpmethod: methodname for httpmethod in httpmethods},
                    name=replace_methodname(route.name, url_path),
                    initkwargs=initkwargs,
                ))

            return ret

        ret = []

        for route in self.routes:
            if isinstance(route, DynamicDetailRoute):
                # Dynamic detail routes (@detail_route decorator)
                ret += _get_dynamic_routes(route, detail_routes)
            elif isinstance(route, DynamicListRoute):
                # Dynamic list routes (@list_route decorator)
                ret += _get_dynamic_routes(route, list_routes)
            else:
                # Standard route
                ret.append(route)

        return ret

    def get_lookup(self, viewset):
        base_regex = '{%s}'
        lookup_field = getattr(viewset, 'lookup_field', 'id')
        lookup_url_kwargs = getattr(viewset, 'lookup_url_kwargs', None)

        if lookup_url_kwargs:
            lookup_url_keys = list(lookup_url_kwargs)

            if len(lookup_url_keys) > 1:
                raise ImproperlyConfigured('ViewSetRouter does not support nested routes.')

            lookup_url_kwarg = lookup_url_keys[0]
        else:
            lookup_url_kwarg = lookup_field

        return base_regex % lookup_url_kwarg

    def get_method_map(self, viewset, method_map):
        """
        Given a viewset, and a mapping of http methods to actions, return a new mapping which only
        includes any mappings that are actually implemented by the viewset.
        """

        bound_methods = {}

        for method, action in method_map.items():
            if hasattr(viewset, action):
                bound_methods[method] = action

        return bound_methods



from marshmallow import SchemaOpts, pre_dump

__all__ = ['ExpandableSchemaMixin',
           'ExpandableViewMixin',
           'ExpandableOpts']


def parse_requested_expands(query_key, request):
    """
    Extracts the value of the expand query string parameter from a request.
    Supports comma separated lists.

    :param query_key: The name query string parameter.
    :param request: Request instance.
    :return: List of strings representing the values of the expand query string value.
    """

    requested_expands = []

    for key, val in request.params.items():
        if key == query_key:
            requested_expands += val.split(',')

    return requested_expands


class ExpandableOpts(SchemaOpts):
    """
    Adds support for expandable_fields to Class Meta. `expandable_fields` should be a
    dict of key = field name to be added to serialized results and val = a marshmallow Nested field.

    Example:
        expandable_fields = dict(paymentmethod=fields.Nested(PaymentMethodSchema, required=False))
    """

    def __init__(self, meta):
        super(ExpandableOpts, self).__init__(meta)
        self.expandable_fields = getattr(meta, 'expandable_fields', dict())


class ExpandableSchemaMixin:
    """
    Supports optionally expandable fields based on QUERY_KEY query strings included in the request's url.
    """

    OPTIONS_CLASS = ExpandableOpts
    QUERY_KEY = 'expand'

    @pre_dump
    def update_expandables(self, data):
        request = self.context.get('request')

        if request:
            requested_expands = parse_requested_expands(self.QUERY_KEY, request)
            available_expands = self.opts.expandable_fields.keys()

            for field in requested_expands:
                if field in available_expands:
                    self.declared_fields[field] = self.opts.expandable_fields[field]

        return data


class ExpandableViewMixin:
    """
    Optionally used to allow more fine grained control over the query used to pull data.
    `expandable_fields` should be a dict of key = the field name that is expandable
    and val = is a dict with the following keys:

    join (optional):        A table column to join() to the query.
    outerjoin (optional):   A table column to outerjoin() to the query.
    options (optional):     list passed to the constructed queries' options method. This is where you
                            want to include the related objects to expand on. Without a value you here
                            you will likely end up running lots of extra queries.

    Example:
        expandable_fields = {
            'author': {'column': Book.author, 'options': [joinedload(Book.author)]
        }
    """

    expandable_fields = None

    def get_query(self):
        """
        If your query is more complicated than what is supported below, override this method.
        Don't forget to call super though.
        """

        query = super(ExpandableViewMixin, self).get_query()
        expandable_fields = getattr(self, 'expandable_fields', [])

        if expandable_fields:
            requested_expands = parse_requested_expands(self.schema_class.QUERY_KEY, self.request)

            if requested_expands:
                available_expands = self.expandable_fields.keys()

                for name in requested_expands:
                    if name in available_expands:
                        field = self.expandable_fields[name]

                        innerjoin = field.get('join')
                        outerjoin = field.get('outerjoin')

                        if innerjoin:
                            query = query.join(innerjoin)
                        elif outerjoin:
                            query = query.outerjoin(outerjoin)

                        # Apply optional options
                        options = field.get('options')

                        if options:
                            query = query.options(*options)

        return query



from sqlalchemy import or_, ARRAY, func


class BaseFilter:
    """
    Base filter that all filter classes must implement.
    """

    def filter_query(self, request, query, view):
        """
        Return the filter query.
        """
        raise NotImplementedError('.filter_query() must be implemented.')  # pragma: no cover


class AttributeBaseFilter(BaseFilter):
    """
    A base class for implementing filters on SQLAlchemy model attributes.
    Supports filtering a comma separated list using OR statements and relationship filter using
    the . path to attribute. WARNING: Every relationship in a . path is joined.

    query_string_lookup: The key to use when parsing the requests query string. The <key> in <key>[<field_name>]=<val>.
    view_attribute_name: The attribute name used in the view that uses the filter that specifies which attributes
                         can be filtered on.
    """

    query_string_lookup = None
    view_attribute_name = None

    def parse_query_string(self, params):
        """
        Override this method if you need to support query string filter keys other than those in the
        format of filter[<field_name>].

        :return: dict
        """
        results = {}

        for key, val in params.items():
            lookup_len = len(self.query_string_lookup) + 1

            if key[0:lookup_len] == '{}['.format(self.query_string_lookup) and key[-1] == ']':
                results[key[lookup_len:-1]] = val

        return results

    def filter_query(self, request, query, view):
        if not request.params:
            return query

        querystring_params = self.parse_query_string(request.params)
        query, filter_list = self.build_filter_list(querystring_params, query, view)

        return self.apply_filter(query, filter_list)

    def build_filter_list(self, querystring_params, query, view):
        filterable_fields = getattr(view, self.view_attribute_name, None)

        if not filterable_fields:
            return query, []

        filter_list = []
        available_fields = list(map(lambda x: '{}.{}'.format(x.parent.class_.__name__, x.name), filterable_fields))

        for key, val in querystring_params.items():
            attrs = key.split('.')  # Every item in the resulting array must be a relationship except for the last
            related_model = view.model
            join_models = []

            for attr in attrs[:-1]:
                # Loop through all the relationships and build the list of tables that need to be joined
                # to the query. The final value of related_model will be the model representing the table
                # with the attribute to be filtered on.

                relationship = getattr(related_model, attr, None)

                if relationship is None:
                    related_model = None
                    break

                related_model = relationship.mapper.class_
                join_models.append(related_model)

            attr = attrs[-1]

            if related_model and hasattr(related_model, attr):
                try:
                    i = available_fields.index('{}.{}'.format(related_model.__name__, attr))
                except ValueError:
                    continue

                joined_tables = [mapper.class_ for mapper in query._join_entities]  # accessing protected attribute?

                for join_model in join_models:
                    if join_model not in joined_tables:
                        query = query.join(join_model)

                filter_list.append(self.build_comparision(filterable_fields[i], val))

        return query, filter_list

    def apply_filter(self, query, filter_list):
        """
        Override this if you need to do something beside calling filter.

        :param query: the query that will be returned from the filter_query method.
        :param filter_list: An array of SQLAlchemy comparative statements.
        :return: The query
        """

        return query.filter(*filter_list)

    def build_comparision(self, field, value):
        """
        Given the model field and the value to be filtered, this should return the statement to be appended
        as a filter to the final query.
        """

        raise NotImplementedError


class FieldFilter(AttributeBaseFilter):
    """
    Filters a query based on the filter_fields set on the view.
    filter_fields should be a list of SQLAlchemy Model columns.

    Comma separated values are treated as ORs. Multiple filter[<field>] query params are AND'd together.
    """

    query_string_lookup = 'filter'
    view_attribute_name = 'filter_fields'

    def build_comparision(self, field, value):
        # Support "IN" filtering
        return or_(*[field == v for v in value.split(',')])


class SearchFilter(AttributeBaseFilter):
    """
    Implements LIKE filtering based on the search[<field_name>]=<val> querystring.
    Comma separated values are treated as ORs.
    Multiple search[<fields>] are or'd together.
    """

    query_string_lookup = 'search'
    view_attribute_name = 'search_fields'

    def build_comparision(self, field, value):
        if issubclass(field.type.__class__, ARRAY):
            return or_(*[field.any(v.lower()) for v in value.split(',')])

        return or_(*[func.lower(field).like('%{}%'.format(v.lower())) for v in value.split(',')])

    def apply_filter(self, query, filter_list):
        return query.filter(or_(*filter_list))


class OrderFilter(AttributeBaseFilter):
    """"
    Allow ordering of the query based on an order[<field>]=<asc || desc> query string.
    """

    query_string_lookup = 'order'
    view_attribute_name = 'order_fields'

    def build_comparision(self, field, value):
        return field if value != 'desc' else field.desc()

    def apply_filter(self, query, filter_list):
        return query.order_by(*filter_list)



def detail_route(methods=None, **kwargs):
    """
    Used to mark a method on a ViewSet that should be routed for detail requests.

    :param methods: An iterable of strings representing the HTTP (GET, POST, etc.) methods accepted by the route.
    :param url_path: Replaces the route automatically generated by the ViewSetRouter for the decorated method
                     with the value provided.
    """
    methods = ['get'] if (methods is None) else methods

    def decorator(func):
        func.bind_to_methods = methods
        func.detail = True
        func.kwargs = kwargs

        return func

    return decorator


def list_route(methods=None, **kwargs):
    """
    Used to mark a method on a ViewSet that should be routed for list requests.

    :param methods: An iterable of strings representing the HTTP (GET, POST, etc.) methods accepted by the route.
    :param url_path: Replaces the route automatically generated by the ViewSetRouter for the decorated method
                     with the value provided.
    """

    methods = ['get'] if (methods is None) else methods

    def decorator(func):
        func.bind_to_methods = methods
        func.detail = False
        func.kwargs = kwargs

        return func

    return decorator



import math

from pyramid.response import Response

from .pagenumber import PageNumberPagination
from .utilities import replace_query_param


__all__ = ['LinkHeaderPagination']


class LinkHeaderPagination(PageNumberPagination):
    """
    DOC GOES HERE
    """

    def get_paginated_response(self, data):
        next_url = self.get_next_link()
        previous_url = self.get_previous_link()
        first_url = self.get_first_link()
        last_url = self.get_last_link()

        link = ''

        if next_url is not None and previous_url is not None:
            link = '<{next_url}>; rel="next", <{previous_url}>; rel="prev"'
        elif next_url is not None:
            link = '<{next_url}>; rel="next"'
        elif previous_url is not None:
            link = '<{previous_url}>; rel="prev"'

        if link:
            link += ', <{first_url}>; rel="first", <{last_url}>; rel="last"'

        response = Response(json=data)  # todo, support renderer, should not hard code json
        link = link.format(next_url=next_url, previous_url=previous_url, first_url=first_url, last_url=last_url)

        if link:
            response.headers['Link'] = link
            response.headers['X-Total-Count'] = str(self.page.paginator.count)

        return response

    def get_first_link(self):
        url = self.get_url_root()
        return replace_query_param(url, self.page_query_param, 1)

    def get_last_link(self):
        url = self.get_url_root()
        count = self.page.paginator.count
        page_size = self.get_page_size(self.request)

        total_pages = int(math.ceil(count / float(page_size)))

        return replace_query_param(url, self.page_query_param, total_pages)



from pyramid.compat import urlparse


def replace_query_param(url, key, val):
    """
    Given a URL and a key/val pair, set or replace an item in the query
    parameters of the URL, and return the new URL.
    """

    (scheme, netloc, path, query, fragment) = urlparse.urlsplit(url)
    query_dict = urlparse.parse_qs(query)
    query_dict[key] = [val]
    query = urlparse.urlencode(sorted(list(query_dict.items())), doseq=True)
    return urlparse.urlunsplit((scheme, netloc, path, query, fragment))


def remove_query_param(url, key):
    """
    Given a URL and a key/val pair, remove an item in the query
    parameters of the URL, and return the new URL.
    """

    (scheme, netloc, path, query, fragment) = urlparse.urlsplit(url)
    query_dict = urlparse.parse_qs(query)
    query_dict.pop(key, None)
    query = urlparse.urlencode(sorted(list(query_dict.items())), doseq=True)
    return urlparse.urlunsplit((scheme, netloc, path, query, fragment))
